import processing.core.*; 
import processing.data.*; 
import processing.event.*; 
import processing.opengl.*; 

import ddf.minim.*; 
import ddf.minim.signals.*; 
import ddf.minim.ugens.*; 

import java.util.HashMap; 
import java.util.ArrayList; 
import java.io.File; 
import java.io.BufferedReader; 
import java.io.PrintWriter; 
import java.io.InputStream; 
import java.io.OutputStream; 
import java.io.IOException; 

public class FoxDance2 extends PApplet {

// import minim libraries:




// Tweakable game variables:
int beat_length = 750;
float base_volume = 0.02f;
int keyAlarm_length = 100;
int mouseAlarm_length = 100;
float grid_y = 280; // y -offset of the grid
int rows = 9;
int columns = 16;
float tile_size = 768/columns;

// mod defines number of octaves
// set to 12 or 24 for 1 or 2 octaves
// if 12, all notes fall in one octave but some chords will be inverted
// if 24, no chords will be inverted but may have large jumps between some chords
int mod=12;

// For noise
float time_offset = 0.0f;

// Frequencies of various notes
float[] notes= {
  261.63f, //C
  277.18f, 
  293.66f, //D
  311.13f, 
  329.63f, //E
  349.23f, //F
  369.99f, 
  392.0f, //G
  415.3f, 
  440.0f, //A
  466.16f, 
  493.88f, //B
  523.25f, 
  554.37f, 
  587.33f, 
  622.25f, 
  659.26f, 
  698.46f, 
  739.99f, 
  783.99f, 
  830.61f, 
  880.00f, 
  932.33f, 
  987.77f
};

// For key presses
boolean up_pressed, down_pressed, left_pressed, right_pressed;
boolean up_right, up_left, down_right, down_left;

// Global objects
Minim minim;
Fox fox1, fox2, fox3;
PImage foxLeft, foxRight, background;


public void setup()
{
  size (768, 768);
  println(tile_size);

  minim = new Minim(this);
  fox1 = new Fox( 7, 12, '1', color(247, 119, 45) );
  fox2 = new Fox( 9, 3, '2', color(152, 178, 52) );
  fox3 = new Fox( 2, 6, '3', color(223, 31, 66) );
  foxLeft = loadImage("foxLeft.png");
  foxRight = loadImage("foxRight.png");
  background = loadImage("background.png");
}


public void draw() 
{ 
  // Check for diagonal key presses
  up_right = false;
  up_left = false;
  down_right = false;
  down_left = false;
  if (up_pressed && right_pressed) up_right = true;
  if (up_pressed && left_pressed) up_left = true;
  if (down_pressed && right_pressed) down_right = true;
  if (down_pressed && left_pressed) down_left = true;

  fox1.update();
  fox2.update();
  fox3.update();

  background(10, 25, 56);
  fill(10, 25, 56);
  noStroke();
  rect(0, 0, 400, 200);

  noTint();
  image(background, 0, 0);

  // draw tile grid
  noStroke();
  float x_offset = 0.0f;
  for (int r = 0; r < rows; r++) {
    float y_offset = 0.0f;
    for (int c = 0; c < columns; c++) {
      int row=12-r;
      int col=c;
      fill( 250, 235, 194, map(noise(x_offset, y_offset, time_offset), 0, 1, 189, 246));
      rect(c*tile_size, grid_y+(tile_size*r), tile_size, tile_size);
      y_offset += 0.5f;
    }
    x_offset += 0.5f;
  }

  fox1.display(10, height-50);
  fox2.display(50, height-50);
  fox3.display(90, height-50);

  // Display text
  textSize(13);
  fill(198);
  text("Controls:", 175, height-35);
  text("1, 2 & 3: Take or give up control of a fox", 190, height-20);
  text("4: Take control of all foxes", 190, height-7);
  text("Spacebar: Give up control of all foxes", 475, height-35);
  text("Arrow keys: Move controlled foxes", 475, height-20);
  text("Q, W, E, A, S, & D: Change chords", 475, height-7);

  time_offset += 0.05f;
}


public void stop()
{
  // always close Minim audio classes when you are done with them
  fox1.closeAll();
  super.stop();
}

class Alarm
{
  float target_millis;
  boolean finished;


  Alarm(float target_millis_)
  {
    target_millis = target_millis_ + millis();
  }


  public void run()
  {
    if (millis() >= target_millis) finished = true;
    else finished = false;
  }
}
class Fox
{
  // audio outputs for 3 notes
  AudioOutput out1;
  AudioOutput out2;
  AudioOutput out3;
  AudioOutput out4;

  Oscil wave1, wave2, wave3, wave4;

  // keyAlarm2 is used for number keys
  Alarm keyAlarm, keyAlarm2, mouseAlarm;

  // notes being played 0..1 eg. C major will be 0,4,7
  int note1=0;
  int note2=0;
  int note3=0;
  int note4=0;

  float volume;

  // Location on the tile grid
  int column, row;

  int current_chord;

  int move_index;

  // Whether or not this fox is currently being controlled by the player
  boolean active;

  // Colors:
  int rootColor, noteColor;

  // Key which is pressed to toggle active on or off
  char toggle_key;

  int time_of_last_keypress;

  PImage currentSprite, activeIcon;

  ArrayList<SavedMove> savedMoves = new ArrayList<SavedMove>();


  Fox(int column_, int row_, char toggle_key_, int rootColor_)
  {
    column = column_;
    row = row_;

    toggle_key = toggle_key_;

    volume = 0;

    // Get a line out from Minim, 
    // default sample rate is 44100, default bit depth is 16
    out1 = minim.getLineOut(Minim.STEREO, 1024);
    out2 = minim.getLineOut(Minim.STEREO, 1024);
    out3 = minim.getLineOut(Minim.STEREO, 1024);
    out4 = minim.getLineOut(Minim.STEREO, 1024);

    out1.setPan(random(-1, 1));
    out2.setPan(random(-1, 1));
    out3.setPan(random(-1, 1));
    out4.setPan(random(-1, 1));

    // create a sine wave Oscillator, 
    // start with a single note
    // choose a random waveform
    float startingNote = notes[PApplet.parseInt(random(24))];
    wave1 = new Oscil(startingNote, 0.25f, Waves.SINE);
    wave2 = new Oscil(0, 0.25f, Waves.SINE);
    wave3 = new Oscil(0, 0.25f, Waves.SINE);
    wave4 = new Oscil(0, 0.25f, Waves.SINE);
    setRandomWaveform(wave1);
    setRandomWaveform(wave2);
    setRandomWaveform(wave3);
    setRandomWaveform(wave4);

    wave1.setAmplitude(volume);
    wave2.setAmplitude(volume);
    wave3.setAmplitude(volume);
    wave4.setAmplitude(volume);

    // add the oscillator to the line out
    wave1.patch(out1);
    wave2.patch(out2);
    wave3.patch(out3);
    wave4.patch(out4);

    active = false;

    // Initialize color
    rootColor = color( red(rootColor_), green(rootColor_), blue(rootColor_), 225);
    noteColor = color( red(rootColor), green(rootColor), blue(rootColor), 175);

    keyAlarm = new Alarm(keyAlarm_length);
    keyAlarm2 = new Alarm(keyAlarm_length);
    mouseAlarm = new Alarm(mouseAlarm_length);

    currentSprite = loadImage("foxLeft.png");
    activeIcon = loadImage("foxLeft.png");

    current_chord = 4;
    // Change chord
    int tonic_key=((7*column)+(4*row))%12;
    changeChord(tonic_key, current_chord);

    move_index = 0;

    time_of_last_keypress = 0;
  }


  public void update()
  { 
    keyAlarm.run();
    keyAlarm2.run();
    mouseAlarm.run();

    if (active) {
      // Check for active toggle
      if (keyPressed == true && keyAlarm2.finished && (key==toggle_key || key==' ')) {
        active = false;
        keyAlarm2 = new Alarm(keyAlarm_length);
      }

      if (mousePressed && mouseAlarm.finished) {
        if (mouseY < grid_y || mouseY > (grid_y+tile_size*rows)) return; // only click on grid!
        // change the tonic (key note) of chord depending on which square you clicked
        println("did it");
        // Get the appropriate row & column
        int target_row = PApplet.parseInt( ((mouseY-grid_y)/tile_size) );
        int target_column = PApplet.parseInt( mouseX/tile_size );

        println(target_row+", "+target_column);

        // Change sprite
        if ( target_column > column ) currentSprite = loadImage("foxRight.png");
        else if (target_column < column) currentSprite = loadImage("foxLeft.png");

        // Save move
        savedMoves.add(new SavedMove(target_row, target_column, current_chord, millis()-time_of_last_keypress));
        time_of_last_keypress = millis();

        row = target_row;
        column = target_column;

        // Change chord
        int tonic_key=((7*column)+(4*row))%12;
        changeChord(tonic_key, current_chord);

        mouseAlarm = new Alarm(mouseAlarm_length);
      }


      if (keyPressed == true && keyAlarm.finished) {
        boolean change_chord = false;
        if (up_right == true) {
          row -= 1;
          column += 1;
          change_chord = true;
        } else if (up_left == true) {
          row -= 1;
          column -= 1;
          change_chord = true;
        } else if (down_right == true) {
          row += 1;
          column += 1;
          change_chord = true;
        } else if (down_left == true) {
          row += 1;
          column -= 1;
          change_chord = true;
        } else if (key == CODED ) {
          if (keyCode == UP) {
            row -= 1;
            change_chord = true;
          } else if (keyCode == DOWN) {
            row += 1;
            change_chord = true;
          } else if (keyCode == LEFT) {
            column -= 1;
            change_chord = true;
            currentSprite = loadImage("foxLeft.png");
          } else if (keyCode == RIGHT) {
            column += 1;
            change_chord = true;
            currentSprite = loadImage("foxRight.png");
          }
        } else if (key != CODED) {
          if (key == 'q') {
            current_chord = 0;
            change_chord = true;
          } else if (key=='w') {
            current_chord = 1;
            change_chord = true;
          } else if (key=='e') {
            current_chord = 2;
            change_chord = true;
          } else if (key=='a') {
            current_chord = 3;
            change_chord = true;
          } else if (key=='s') {
            current_chord = 4;
            change_chord = true;
          } else if (key=='d') {
            current_chord = 5;
            change_chord = true;
          }
        }
        wrap();
        if (change_chord == true) {
          // Save last move
          savedMoves.add(new SavedMove(row, column, current_chord, millis()-time_of_last_keypress));
          time_of_last_keypress = millis();

          // Change chord
          int tonic_key=((7*column)+(4*row))%12;
          changeChord(tonic_key, current_chord);
        }

        keyAlarm = new Alarm( keyAlarm_length );
      }
    } else if (!active) {
      // Check for toggle key
      if (keyPressed == true && keyAlarm2.finished && (key==toggle_key || key=='4')) {
        savedMoves.clear();
        time_of_last_keypress = millis();
        move_index = 0;
        keyAlarm2 = new Alarm(keyAlarm_length);
        active = true;
      }

      if (savedMoves.size() > 0) {
        SavedMove sm = savedMoves.get(move_index);
        sm.update();
        if (sm.ready) {
          if (sm.column > column) currentSprite = loadImage("foxRight.png");
          else if (sm.column < column) currentSprite = loadImage("foxLeft.png");
          row = sm.row;
          column = sm.column;
          wrap();
          current_chord = sm.chord;

          sm.ready = false;
          sm.initialized = false;
          move_index += 1;

          if (move_index >= savedMoves.size()) {
            move_index = 0;
          }

          // Change chord
          int tonic_key=((7*column)+(4*row))%12;
          changeChord(tonic_key, current_chord);
        }
      }
    }

    wave1.setAmplitude(volume);
    wave2.setAmplitude(volume);
    wave3.setAmplitude(volume);
    wave4.setAmplitude(volume);

    // Handle decay
    volume = max(0, volume*0.93f);
  }



  public void display(float x_, float y_)
  {
    // draw the waveform
    // simply add the three waveforms together and multiply a bit
    stroke(0, 0, 255, 255);
    for (int i = 0; i < out1.left.size ()-1; i++)
    {
      float amp1 = 100.0f+((out1.left.get(i))*1000) + ((out2.left.get(i))*1000) + ((out3.left.get(i))*1000);
      float amp2 = 100.0f+((out1.left.get(i+1))*1000) + ((out2.left.get(i+1))*1000) + ((out3.left.get(i+1))*1000);
      line((float)i, amp1+40, (float)(i+1), amp2+40);
    }

    // Draw the foxes themselves
    int rootColor_ = rootColor;
    int noteColor_ = noteColor;
    if (!active) {
      rootColor_ = color( red(rootColor), green(rootColor), blue(rootColor), 125 );
      noteColor_ = color( red(rootColor_), green(rootColor_), blue(rootColor_), 75);
    }

    for (int r = 0; r < rows; r++) {
      for (int c = 0; c < columns; c++) {

        int tonic_key = ( (7*c)+(4*r) )%12;
        boolean display = false;

        int note = 0;

        if ( tonic_key == note2%12 || tonic_key == note3%12 || tonic_key == note4%12 ) {
          tint(noteColor_);
          display = true;
        }
        if ( tonic_key == note1%mod ) {
          tint(rootColor_);
          display = true;
        }

        if (display == true) {
          image(currentSprite, c*tile_size, grid_y+(tile_size*r), tile_size, tile_size);
        }
      }
    }

    tint(rootColor_); 
    // Display active icon
    image(activeIcon, x_, y_);
  }


  public void changeChord(int key_number, int type)
  {
    // change chord
    // key_number is root (0=C,1=C#...11=B)
    //
    // note that currently restricted to one octave, so higher key notes like AMinor, B Major
    // will play as inverted chords. These may not sound as good as the 'usual' uninverted chords

    note1 = key_number;
    note2 = key_number;
    note3 = key_number;
    note4 = key_number;

    // Major key
    if (type == 0) {
      note2 = (key_number+4) % mod;
      note3 = (key_number+7) % mod;
      note4 = (key_number) % mod;
    }

    // Minor key
    if (type == 1) {
      note2 = (key_number+3) % mod;
      note3 = (key_number+7) % mod;
      note4 = (key_number) % mod;
    }

    // Major7
    if (type == 2) {
      note2 = (key_number+4) % mod;
      note3 = (key_number+7) % mod;
      note4 = (key_number+11) % mod;
    }

    // Minor7
    if (type == 3) {
      note2 = (key_number+4) % mod;
      note3 = (key_number+7) % mod;
      note4 = (key_number+10) % mod;
    }

    // Single note
    if (type == 4) {
      note2 = key_number % mod;
      note3 = key_number % mod;
      note4 = key_number % mod;
    }

    // Augmented triad
    if (type == 5) {
      note2 = (key_number+4) % mod;
      note3 = (key_number+8) % mod;
      note4 = key_number % mod;
    }

    // Diminished triad
    if (type == 6) {
      note2 = (key_number+2) % mod;
      note3 = (key_number+6) % mod;
      note4 = key_number % mod;
    }

    note1 = constrain(note1, 0, notes.length);
    note2 = constrain(note2, 0, notes.length);
    note3 = constrain(note3, 0, notes.length);
    note4 = constrain(note4, 0, notes.length);

    wave1.setFrequency(notes[note1]);
    wave2.setFrequency(notes[note2]);
    wave3.setFrequency(notes[note3]);
    wave4.setFrequency(notes[note4]);

    volume = base_volume;
  }

  public void wrap() {
    // Handle wrapping
    row = constrain(row, -1, 3);
    column = constrain(column, -1, mod);
    if (row < 0) {
      row = 2;
    }
    if (row > 2) {
      row = 0;
    }
    if (column < 0) {
      column = mod-1;
    }
    if (column > mod-1) {
      column = 0;
    }
  }


  public void closeAll()
  {
    // always close Minim audio classes when you are done with them
    out1.close();
    out2.close();
    out3.close();
    out4.close();
  }
}

class SavedMove
{
  // Time since last action
  int intended_millis;
  int starting_millis;
  
  // Intended move
  int row, column, chord;
  
  Alarm myAlarm;
  
  boolean initialized;
  boolean ready;
  
  
  SavedMove(int row_, int column_, int chord_, int intended_millis_)
  {
    intended_millis = intended_millis_;
    
    row = row_;
    column = column_;
    chord = chord_;
    
    starting_millis = millis();
    
    initialized = false;
    
    ready = false;
  }
  
  
  public void update() {
    if (initialized == false) {
      starting_millis = millis();
      initialized = true;
    }
    
    if (millis() - starting_millis > intended_millis) {
      ready = true;
    }
  }
}
class ArrowKeys {
  
}
public void setRandomWaveform(Oscil osil_) {
  float r = random(10);
  if (r < 1) osil_.setWaveform(Waves.SINE);
  else if (r < 2) osil_.setWaveform(Waves.triangle(random(0,1)));
  else if (r < 3) osil_.setWaveform(Waves.saw(random(0,1)));
  else if (r < 4) osil_.setWaveform(Waves.square(random(0,1)));
  else if (r < 5) osil_.setWaveform(Waves.QUARTERPULSE);
  else if (r < 6) osil_.setWaveform(Waves.PHASOR);
  else if (r < 7) osil_.setWaveform(Waves.randomNoise());
  else if (r < 8) osil_.setWaveform(Waves.pulse(random(1)));
  else if (r < 9) osil_.setWaveform(Waves.randomNHarms(PApplet.parseInt(random(3))));
  else if (r < 10) osil_.setWaveform(Waves.randomNOddHarms(PApplet.parseInt(random(3))));  
}

public void keyPressed() {
  if (key == CODED) {
    if (keyCode == UP) up_pressed = true;
    else if (keyCode == DOWN) down_pressed = true;
    else if (keyCode == LEFT) left_pressed = true;
    else if (keyCode == RIGHT) right_pressed = true;
  }
}


public void keyReleased() {
  if (key == CODED) {
    if (keyCode == UP) up_pressed = false;
    else if (keyCode == DOWN) down_pressed = false;
    else if (keyCode == LEFT) left_pressed = false;
    else if (keyCode == RIGHT) right_pressed = false;
  }
}

  static public void main(String[] passedArgs) {
    String[] appletArgs = new String[] { "FoxDance2" };
    if (passedArgs != null) {
      PApplet.main(concat(appletArgs, passedArgs));
    } else {
      PApplet.main(appletArgs);
    }
  }
}
