// import minim libraries:
import ddf.minim.*;
import ddf.minim.signals.*;
import ddf.minim.ugens.*;

// Tweakable game variables:
int beat_length = 750;
float base_volume = 0.02;
int keyAlarm_length = 100;
int mouseAlarm_length = 100;
float grid_y = 280; // y -offset of the grid
int rows = 9;
int columns = 16;
float tile_size = 768/columns;

// mod defines number of octaves
// set to 12 or 24 for 1 or 2 octaves
// if 12, all notes fall in one octave but some chords will be inverted
// if 24, no chords will be inverted but may have large jumps between some chords
int mod=12;

// For noise
float time_offset = 0.0;

// Frequencies of various notes
float[] notes= {
  261.63f, //C
  277.18f, 
  293.66f, //D
  311.13f, 
  329.63f, //E
  349.23f, //F
  369.99f, 
  392.0f, //G
  415.3f, 
  440.0f, //A
  466.16f, 
  493.88f, //B
  523.25f, 
  554.37f, 
  587.33f, 
  622.25f, 
  659.26f, 
  698.46f, 
  739.99f, 
  783.99f, 
  830.61f, 
  880.00f, 
  932.33f, 
  987.77f
};

// For key presses
boolean up_pressed, down_pressed, left_pressed, right_pressed;
boolean up_right, up_left, down_right, down_left;

// Global objects
Minim minim;
Fox fox1, fox2, fox3;
PImage foxLeft, foxRight, background;


void setup()
{
  size (768, 768);
  println(tile_size);

  minim = new Minim(this);
  fox1 = new Fox( 7, 12, '1', color(247, 119, 45) );
  fox2 = new Fox( 9, 3, '2', color(152, 178, 52) );
  fox3 = new Fox( 2, 6, '3', color(223, 31, 66) );
  foxLeft = loadImage("foxLeft.png");
  foxRight = loadImage("foxRight.png");
  background = loadImage("background.png");
}


void draw() 
{ 
  // Check for diagonal key presses
  up_right = false;
  up_left = false;
  down_right = false;
  down_left = false;
  if (up_pressed && right_pressed) up_right = true;
  if (up_pressed && left_pressed) up_left = true;
  if (down_pressed && right_pressed) down_right = true;
  if (down_pressed && left_pressed) down_left = true;

  fox1.update();
  fox2.update();
  fox3.update();

  background(10, 25, 56);
  fill(10, 25, 56);
  noStroke();
  rect(0, 0, 400, 200);

  noTint();
  image(background, 0, 0);

  // draw tile grid
  noStroke();
  float x_offset = 0.0;
  for (int r = 0; r < rows; r++) {
    float y_offset = 0.0;
    for (int c = 0; c < columns; c++) {
      int row=12-r;
      int col=c;
      fill( 250, 235, 194, map(noise(x_offset, y_offset, time_offset), 0, 1, 189, 246));
      rect(c*tile_size, grid_y+(tile_size*r), tile_size, tile_size);
      y_offset += 0.5;
    }
    x_offset += 0.5;
  }

  fox1.display(10, height-50);
  fox2.display(50, height-50);
  fox3.display(90, height-50);

  // Display text
  textSize(13);
  fill(198);
  text("Controls:", 175, height-35);
  text("1, 2 & 3: Take or give up control of a fox", 190, height-20);
  text("4: Take control of all foxes", 190, height-7);
  text("Spacebar: Give up control of all foxes", 475, height-35);
  text("Arrow keys: Move controlled foxes", 475, height-20);
  text("Q, W, E, A, S, & D: Change chords", 475, height-7);

  time_offset += 0.05;
}


void stop()
{
  // always close Minim audio classes when you are done with them
  fox1.closeAll();
  super.stop();
}

